"""Module level docstring.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from pathlib import Path
from typing import List, Set
from tqdm import tqdm
from digiarch.internals import FileInfo, IGNORED_EXTS

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def grouping(files: List[FileInfo], save_path: Path) -> None:
    """Groups files per file extension.

    Parameters
    ----------
    data_file : str
        File from which data is read.
    save_path : str
        Path to save results to
    """

    # Initialise variables
    ignored_group: List[Path] = []
    exts: Set[str] = {file.ext for file in files}

    # Create new folder in save path
    save_path = save_path / "grouped_files"
    save_path.mkdir(exist_ok=True)

    # Group files per file extension.
    for ext in tqdm(exts, desc="Grouping files"):
        group_out: Path = save_path / f"{ext}_files.txt"
        file_group: List[Path] = []
        for file in files:
            if file.ext in IGNORED_EXTS:
                ignored_group.append(file.path)
            if file.ext == ext:
                file_group.append(file.path)
        for file_path in file_group:
            group_out.write_text(f"{file_path}\n", encoding="utf-8")

    # Save ignored files
    ignored_out: Path = save_path / "ignored_files.txt"
    for file_path in ignored_group:
        ignored_out.write_text(f"{file_path}\n", encoding="utf-8")



"""Implements data classes and related utilities used throughout
Digital Archive.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import dataclasses
import inspect
import json
import dacite
from datetime import datetime
from dateutil.parser import parse as date_parse
from pathlib import Path
from typing import Any, List, Optional, Set
from natsort import natsorted
import digiarch

# -----------------------------------------------------------------------------
# Globals
# -----------------------------------------------------------------------------

IGNORED_EXTS: Set[str] = json.load(
    (
        Path(inspect.getfile(digiarch)).parent / "_data" / "blacklist.json"
    ).open()
).keys()

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

# Base class
# --------------------


@dataclasses.dataclass
class DataBase:
    def to_dict(self) -> dict:
        return dataclasses.asdict(self)

    def replace(self, **fields: Any) -> Any:
        return dataclasses.replace(self, **fields)

    @classmethod
    def from_dict(cls, data: dict) -> Any:
        # For some reason the post init yields str for FileData
        if isinstance(data.get("digiarch_dir"), str):
            data.update({"digiarch_dir": Path(data.get("digiarch_dir", ""))})
        if isinstance(data.get("json_file"), str):
            data.update({"json_file": Path(data.get("json_file", ""))})
        return dacite.from_dict(
            data_class=cls, data=data, config=dacite.Config(check_types=False)
        )


# Identification
# --------------------


@dataclasses.dataclass
class Identification(DataBase):
    """Data class for keeping track of file identification information"""

    puid: Optional[str]
    signame: Optional[str]
    warning: Optional[str] = None


# File Info
# --------------------


@dataclasses.dataclass
class FileInfo(DataBase):
    """Data class for keeping track of file information"""

    path: Path
    name: str = dataclasses.field(init=False)
    ext: str = dataclasses.field(init=False)
    size: str = dataclasses.field(init=False)
    checksum: Optional[str] = None
    identification: Optional[Identification] = None

    def __post_init__(self) -> None:
        # JSON/from_dict compatibility
        if not isinstance(self.path, Path):
            self.path = Path(self.path)

        # Resolve path, init fields
        self.path = self.path.resolve()
        self.name = self.path.name
        self.ext = self.path.suffix.lower()
        self.size = size_fmt(self.path.stat().st_size)


# Metadata
# --------------------


@dataclasses.dataclass
class Metadata(DataBase):
    """Data class for keeping track of metadata used in data.json"""

    last_run: datetime
    processed_dir: Path
    file_count: Optional[int] = None
    total_size: Optional[str] = None
    duplicates: Optional[int] = None
    identification_warnings: Optional[int] = None
    empty_subdirs: Optional[List[Path]] = None
    several_files: Optional[List[Path]] = None

    def __post_init__(self) -> None:
        # JSON/from_dict compatibility
        if isinstance(self.processed_dir, str):
            self.processed_dir = Path(self.processed_dir)
        if isinstance(self.last_run, str):
            self.last_run = date_parse(self.last_run)

        # Resolve path
        self.processed_dir = self.processed_dir.resolve()


# JSON Data
# --------------------
@dataclasses.dataclass
class FileData(DataBase):
    """Data class collecting Metadata and list of FileInfo"""

    metadata: Metadata
    files: List[FileInfo] = dataclasses.field(default_factory=list)
    digiarch_dir: Path = dataclasses.field(init=False)
    json_file: Path = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        # Directory paths
        self.digiarch_dir = Path(self.metadata.processed_dir, "_digiarch")
        data_dir = self.digiarch_dir / ".data"
        self.json_file = data_dir / "data.json"

        # Create directories
        self.digiarch_dir.mkdir(exist_ok=True)
        data_dir.mkdir(exist_ok=True)

        # Create data file if it does not exist
        if not self.json_file.is_file():
            self.json_file.touch()

    def to_json(self, file: Optional[Path] = None) -> None:
        if file is None:
            file = self.json_file
        with file.open("w", encoding="utf-8") as f:
            json.dump(
                self, f, indent=4, cls=DataJSONEncoder, ensure_ascii=False
            )

    @classmethod
    def from_json(cls, data_file: Path) -> Any:
        with data_file.open("r", encoding="utf-8") as file:
            return cls.from_dict(json.load(file))


# Utility
# --------------------


class DataJSONEncoder(json.JSONEncoder):
    """DataJSONEncoder subclasses JSONEncoder in order to handle
    encoding of data classes."""

    # pylint does not like this subclassing, even though it's the recommended
    # method. So we disable the warnings.

    # pylint: disable=method-hidden,arguments-differ
    def default(self, obj: object) -> Any:
        """Overrides the JSONEncoder default.

        Parameters
        ----------
        obj : object
            Object to encode.
        Returns
        -------
        dataclasses.asdict(obj) : dict
            If the object given is a data class, return it as a dict.
        super().default(obj) : Any
            If the object is not a data class, use JSONEncoder's default and
            let it handle any exception that might occur.
        """
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        if isinstance(obj, Path):
            return str(obj)
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)

    # pylint: enable=method-hidden,arguments-differ


# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def size_fmt(size: float) -> str:
    """Formats a file size in binary multiples to a human readable string.

    Parameters
    ----------
    size : float
        The file size in bytes.

    Returns
    -------
    str
        Human readable string representing size in binary multiples.
    """
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1024.0:
            break
        size /= 1024.0
    return f"{size:.1f} {unit}"


def to_json(data: object, file: Path) -> None:
    """Dumps JSON files given data and a file path
    using :class:`~digiarch.data.DataJSONEncoder` as encoder.
    Output uses indent = 4 to get pretty and readable files.
    `ensure_ascii` is set to `False` so we can get our beautiful Danish
    letters in the output.

    Parameters
    ----------
    data : object
        The data to dump to the JSON file.
    dump_file: str
        Path to the file in which to dump JSON data.
    """

    with file.open("w", encoding="utf-8") as f:
        json.dump(data, f, indent=4, cls=DataJSONEncoder, ensure_ascii=False)


def natsort_path(file_list: List[FileInfo]) -> List[FileInfo]:
    """Naturally sort a list of FileInfo objects by their paths.

    Parameters
    ----------
    file_list : List[FileInfo]
        The list of FileInfo objects to be sorted.

    Returns
    -------
    List[FileInfo]
        The list of FileInfo objects naturally sorted by their path.
    """

    sorted_file_list: List[FileInfo] = natsorted(
        file_list, key=lambda fileinfo: str(fileinfo.path)
    )

    return sorted_file_list



"""This module implements checksum generation and duplicate detection.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import hashlib
from multiprocessing import Pool
from collections import Counter
from pathlib import Path
from typing import List, Set, Dict, ItemsView, Any, Optional
from tqdm import tqdm
from digiarch.internals import FileInfo, to_json, natsort_path

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def file_checksum(file: Path) -> str:
    """Calculate the checksum of an input file using BLAKE2.

    Parameters
    ----------
    file : Path
        The file for which to calculate the checksum. Expects a `pathlib.Path`
        object.

    Returns
    -------
    str
        The hex checksum of the input file.

    """

    checksum: str = ""
    hasher: Any = hashlib.sha256()

    if file.is_file():
        with file.open("rb") as f:
            hasher.update(f.read())
            checksum = hasher.hexdigest()

    return checksum


def checksum_worker(file_info: FileInfo) -> FileInfo:
    """Worker used when multiprocessing checksums of FileInfo objects.

    Parameters
    ----------
    fileinfo : FileInfo
        The FileInfo object that must be updated with a new checksum value.

    Returns
    -------
    FileInfo
        The FileInfo object with an updated checksum value.
    """

    checksum: Optional[str] = file_checksum(file_info.path) or None
    updated_file_info: FileInfo = file_info.replace(checksum=checksum)
    return updated_file_info


def generate_checksums(files: List[FileInfo]) -> List[FileInfo]:
    """Multiprocesses a list of FileInfo object in order to assign
    new checksums.

    Parameters
    ----------
    files : List[FileInfo]
        List of FileInfo objects that need checksums.

    Returns
    -------
    List[FileInfo]
        The updated list of FileInfo objects.
    """

    # Assign variables
    updated_files: List[FileInfo] = []

    # Multiprocess checksum generation
    pool = Pool()
    try:
        updated_files = list(
            tqdm(
                pool.imap_unordered(checksum_worker, files),
                desc="Generating checksums",
                unit=" files",
                total=len(files),
            )
        )
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()
    finally:
        pool.close()
        pool.join()

    return natsort_path(updated_files)


def check_collisions(checksums: List[str]) -> Set[str]:
    """Checks checksum collisions given a list of checksums as strings.
    Returns a set of collisions if any such are found.

    Parameters
    ----------
    checksums : List[str]
        List of checksums that must be checked for collisions.

    Returns
    -------
    Set[str]
        A set of colliding checksums. Empty if none are found.
    """
    checksum_counts: ItemsView[str, int] = Counter(checksums).items()
    collisions: Set[str] = set()

    for checksum, count in checksum_counts:
        if count > 1:
            # We have a collision, boys
            collisions.add(checksum)

    return collisions


def check_duplicates(files: List[FileInfo], save_path: Path) -> None:
    """Generates a file with checksum collisions, indicating that duplicates
    are present.

    Parameters
    ----------
    files : List[FileInfo]
        Files for which duplicates should be checked.
    save_path : Path
        Path to which the checksum collision information should be saved.
    """

    # Initialise variables
    checksums: List[str] = [
        file.checksum for file in files if file.checksum is not None
    ]
    collisions: Set[str] = check_collisions(checksums)
    file_collisions: Dict[str, str] = dict()

    for checksum in tqdm(collisions, desc="Finding duplicates"):
        hits = [
            {"name": file.name, "path": file.path}
            for file in files
            if file.checksum == checksum
        ]
        file_collisions.update({checksum: hits})

    dups_file = Path(save_path).joinpath("duplicate_files.json")
    to_json(file_collisions, dups_file)



"""Reporting utilities for file discovery.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
from collections import Counter
from pathlib import Path
from typing import List, Dict
from digiarch.internals import FileInfo, Identification, to_json

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def report_results(files: List[FileInfo], save_path: Path) -> None:
    """Generates reports of explore_dir() results.

    Parameters
    ----------
    files: List[FileInfo]
        The files to report on.
    save_path: str
        The path in which to save the reports.

    """

    # Initialise counters & dicts
    ext_count: Counter = Counter()
    id_warnings: Dict[str, List[Dict[str, Identification]]] = dict()
    warning_key: str
    warning_count: Counter = Counter()

    # Collect information
    for file in files:
        ext_count.update([file.ext])
        if file.identification and file.identification.warning:
            if "No match" in file.identification.warning:
                warning_key = "No match"
            else:
                warning_key = file.identification.warning
            warning_list = id_warnings.get(warning_key, [])
            warning_list.append({str(file.path): file.identification})
            id_warnings.update({warning_key: warning_list})

    file_exts: Dict[str, int] = dict(ext_count.most_common())
    if files:
        # Create new folder in save path
        save_path = save_path / "reports"
        save_path.mkdir(exist_ok=True)

        # Save files
        to_json(file_exts, save_path / "file_extensions.json")
        if id_warnings:
            for warning_key, warning_list in id_warnings.items():
                warning_count.update({warning_key: len(warning_list)})

            to_json(
                [dict(warning_count), id_warnings],
                save_path / "identification_warnings.json",
            )



"""Identify files using
`siegfried <https://github.com/richardlehane/siegfried>`_

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import subprocess
from multiprocessing import Pool
from subprocess import CalledProcessError
from typing import List
from digiarch.internals import FileInfo, Identification, natsort_path
from digiarch.exceptions import IdentificationError
import yaml
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def sf_id(file: FileInfo) -> FileInfo:
    """Identify files using
    `siegfried <https://github.com/richardlehane/siegfried>`_ and update
    FileInfo with obtained PUID, signature name, and warning if applicable.

    Parameters
    ----------
    file : FileInfo
        The file to identify.

    Returns
    -------
    updated_file : FileInfo
        Input file with updated information in the Identification field.

    Raises
    ------
    IdentificationError
        If running siegfried or loading of the resulting YAML output fails, an
        IdentificationError is thrown.

    """
    new_id: Identification = Identification(
        puid=None,
        signame=None,
        warning="No identification information obtained.",
    )
    try:
        cmd = subprocess.run(
            ["sf", file.path], capture_output=True, check=True,
        )
    except CalledProcessError as error:
        raise IdentificationError(error)

    try:
        docs = yaml.safe_load_all(cmd.stdout.decode())
    except yaml.YAMLError as error:
        raise IdentificationError(error)
    else:
        match_doc = [doc.get("matches") for doc in docs if "matches" in doc]
        # match_doc is a list of list of matches. Flatten it and get only
        # matches from PRONOM.
        matches = [
            match
            for matches in match_doc
            for match in matches
            if match.get("ns") == "pronom"
        ]
        for match in matches:
            new_id = new_id.replace(
                signame=match.get("format"), warning=match.get("warning")
            )
            if match.get("id", "").lower() == "unknown":
                new_id.puid = None
            else:
                new_id.puid = match.get("id")
            if isinstance(new_id.warning, str):
                new_id.warning = new_id.warning.capitalize()
    updated_file: FileInfo = file.replace(identification=new_id)
    return updated_file


def identify(files: List[FileInfo]) -> List[FileInfo]:
    """Identify all files in a list, and return the updated list.

    Parameters
    ----------
    files : List[FileInfo]
        Files to identify.

    Returns
    -------
    List[FileInfo]
        Input files with updated Identification information.

    """

    updated_files: List[FileInfo]

    # Multiprocess identification
    pool = Pool()
    try:
        updated_files = list(
            tqdm(
                pool.imap_unordered(sf_id, files),
                desc="Identifying files",
                unit="files",
                total=len(files),
            )
        )
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()
    finally:
        pool.close()
        pool.join()

    # Natsort list by file.path
    updated_files = natsort_path(updated_files)

    return updated_files



"""Utilities for handling files, paths, etc.

"""
# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------
import os
import shutil
from datetime import datetime
from pathlib import Path
from tqdm import tqdm
from digiarch.internals import (
    FileInfo,
    Metadata,
    FileData,
    size_fmt,
    natsort_path,
)
from digiarch.exceptions import FileCollectionError
from typing import List

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def explore_dir(path: Path) -> FileData:
    """Finds files and empty directories in the given path,
    and collects them into a list of FileInfo objects.

    Parameters
    ----------
    path : str
        The path in which to find files.

    Returns
    -------
    empty_subs: List[str]
        A list of empty subdirectory paths, if any such were found
    """
    # Type declarations
    dir_info: List[FileInfo] = []
    empty_subs: List[Path] = []
    several_files: List[Path] = []
    total_size: int = 0
    file_count: int = 0
    metadata = Metadata(last_run=datetime.now(), processed_dir=path)
    file_data = FileData(metadata)
    main_dir_name: str = file_data.digiarch_dir.name

    if not [child for child in path.iterdir() if child.name != main_dir_name]:
        # Path is empty, remove main directory and raise
        shutil.rmtree(file_data.digiarch_dir)
        raise FileCollectionError(f"{path} is empty! No files collected.")

    # Traverse given path, collect results.
    # tqdm is used to show progress of os.walk
    for root, dirs, files in tqdm(
        os.walk(path, topdown=True), unit=" folders", desc="Processed"
    ):
        if main_dir_name in dirs:
            # Don't walk the _digiarch directory
            dirs.remove(main_dir_name)
        if not dirs and not files:
            # We found an empty subdirectory.
            empty_subs.append(Path(root))
        if len(files) > 1:
            several_files.append(Path(root))
        for file in files:
            cur_path = Path(root, file)
            dir_info.append(FileInfo(path=cur_path))
            total_size += cur_path.stat().st_size
            file_count += 1

    # Update metadata
    metadata.file_count = file_count
    metadata.total_size = size_fmt(total_size)

    if empty_subs:
        metadata.empty_subdirs = empty_subs
    if several_files:
        metadata.several_files = several_files

    # Update file data
    file_data.files = natsort_path(dir_info)

    # Save file data
    file_data.to_json()

    return file_data
